import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='fleur_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='fleur.nomadmetainfo.json'))


class x_fleur_header(MSection):
    '''
    header (labels) of fleur.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_fleur_header'))

    x_fleur_version = Quantity(
        type=str,
        shape=[],
        description='''
        Version of Fleur
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_version'))


class x_fleur_section_equiv_atoms(MSection):
    '''
    section containing a class of equivalent atoms
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_fleur_section_equiv_atoms'))

    x_fleur_atom_name = Quantity(
        type=str,
        shape=[],
        description='''
        name of atom, labelling non-equvalent atoms
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_atom_name'))

    x_fleur_atom_pos_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        position of atom x
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_atom_pos_x'))

    x_fleur_atom_pos_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        position of atom y
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_atom_pos_y'))

    x_fleur_atom_pos_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        position of atom z
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_atom_pos_z'))

    x_fleur_atom_coord_scale = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        scales coordinates by 1/scale. If film=T, scales only x&y coordinates, if film=F
        also z
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_atom_coord_scale'))

    x_fleur_atomic_number_Z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        atomic number Z
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_atomic_number_Z'))

    x_fleur_nr_equiv_atoms_in_this_atom_type = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        number_equiv_atoms_in_this_atom_type
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_nr_equiv_atoms_in_this_atom_type'))


class x_fleur_section_XC(MSection):
    '''
    exchange-correlation potential
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_fleur_section_XC'))

    x_fleur_exch_pot = Quantity(
        type=str,
        shape=[],
        description='''
        exchange-correlation potential, in out
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_exch_pot'))

    x_fleur_xc_correction = Quantity(
        type=str,
        shape=[],
        description='''
        informaion on relativistic correction for the exchange-correlation potential, in
        out
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_xc_correction'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_fleur_header = SubSection(
        sub_section=SectionProxy('x_fleur_header'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_fleur_header'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_fleur_lattice_vector_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        x component of vector of unit cell
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_lattice_vector_x'))

    x_fleur_lattice_vector_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        y component of vector of unit cell
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_lattice_vector_y'))

    x_fleur_lattice_vector_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        z component of vector of unit cell
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_lattice_vector_z'))

    x_fleur_rec_lattice_vector_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        x component of reciprocal lattice vector
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_rec_lattice_vector_x'))

    x_fleur_rec_lattice_vector_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        y component of reciprocal lattice vector
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_rec_lattice_vector_y'))

    x_fleur_rec_lattice_vector_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        z component of reciprocal lattice vector
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_rec_lattice_vector_z'))

    x_fleur_space_group = Quantity(
        type=str,
        shape=[],
        description='''
        space group
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_space_group'))

    x_fleur_name_of_atom_type = Quantity(
        type=str,
        shape=[],
        description='''
        name of atom type
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_name_of_atom_type'))

    x_fleur_system_nameIn = Quantity(
        type=str,
        shape=[],
        description='''
        user given name for this system given in the inp file
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_system_nameIn'))

    x_fleur_system_name = Quantity(
        type=str,
        shape=[],
        description='''
        user given name for this system
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_system_name'))

    x_fleur_total_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        total number of atoms
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_total_atoms'))

    x_fleur_nr_of_atom_types = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of atom types
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_nr_of_atom_types'))

    x_fleur_nuclear_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        nuclear number
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_nuclear_number'))

    x_fleur_number_of_core_levels = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        x_fleur_number_of_core_levels
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_number_of_core_levels'))

    x_fleur_lexpansion_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        l-expansion cutoff
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_lexpansion_cutoff'))

    x_fleur_mt_gridpoints = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        muffin-tin gridpoints
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_mt_gridpoints'))

    x_fleur_mt_radius = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        muffin-tin radius
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_mt_radius'))

    x_fleur_logarythmic_increment = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        logarythmic increment
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_logarythmic_increment'))

    x_fleur_k_max = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Kmax is the plane wave cut-off
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_k_max'))

    x_fleur_G_max = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Gmax
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_G_max'))

    x_fleur_tot_nucl_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total nuclear charge
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_tot_nucl_charge'))

    x_fleur_tot_elec_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total electronic charge
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_tot_elec_charge'))

    x_fleur_unit_cell_volume = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='bohr ** 3',
        description='''
        unit cell volume
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_unit_cell_volume'))

    x_fleur_unit_cell_volume_omega = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        unit cell volume omega tilda
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_unit_cell_volume_omega'))

    x_fleur_vol_interstitial = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        volume of interstitial region
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_vol_interstitial'))

    x_fleur_section_equiv_atoms = SubSection(
        sub_section=SectionProxy('x_fleur_section_equiv_atoms'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_fleur_section_equiv_atoms'))


class section_scf_iteration(public.section_scf_iteration):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_scf_iteration'))

    x_fleur_tot_for_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TOTAL FORCE FOR ATOM TYPE, X
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_tot_for_x'))

    x_fleur_tot_for_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TOTAL FORCE FOR ATOM TYPE, Y
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_tot_for_y'))

    x_fleur_tot_for_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TOTAL FORCE FOR ATOM TYPE, Z
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_tot_for_z'))

    x_fleur_tot_for_fx = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TOTAL FORCE FOR ATOM TYPE, FX_TOT
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_tot_for_fx'))

    x_fleur_tot_for_fy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TOTAL FORCE FOR ATOM TYPE, FY_TOT
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_tot_for_fy'))

    x_fleur_tot_for_fz = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TOTAL FORCE FOR ATOM TYPE, FZ_TOT
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_tot_for_fz'))

    x_fleur_iteration_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        scf iteration number
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_iteration_number'))

    x_fleur_energy_total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        energy total
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_energy_total'))

    x_fleur_free_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        free energy
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_free_energy'))

    x_fleur_entropy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        (tkb*entropy) TS
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_entropy'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_fleur_nkptd = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of all the k-points
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_nkptd'))

    x_fleur_k_point_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x component of vector of k point
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_k_point_x'))

    x_fleur_k_point_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        y component of vector of k point
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_k_point_y'))

    x_fleur_k_point_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        z component of vector of k point
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_k_point_z'))

    x_fleur_k_point_weight = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        weights of k point
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_k_point_weight'))

    x_fleur_smearing_kind = Quantity(
        type=str,
        shape=[],
        description='''
        The Brillouin zone integration mode. It can be one of hist - Use the histogram
        mode, this is the default; gauss - Use Gaussian smearing, tria - Use the
        tetrahedron method
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_smearing_kind'))

    x_fleur_smearing_width = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        specifies the width of the broadening, smearing for calculation of fermi-energy &
        weights. The Fermi smearing can be parametrized by this energy
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_smearing_width'))

    x_fleur_nr_of_valence_electrons = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The number of electrons to be represented within the valence electron framework
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_nr_of_valence_electrons'))

    x_fleur_smearing_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kelvin',
        description='''
        Fermi smearing temperature set in Kelvin
        ''',
        a_legacy=LegacyDefinition(name='x_fleur_smearing_temperature'))

    x_fleur_section_XC = SubSection(
        sub_section=SectionProxy('x_fleur_section_XC'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_fleur_section_XC'))


m_package.__init_metainfo__()



# pylint: disable=no-self-argument
from typing import Union, List, Optional, Dict, Any

from pydantic import Field, root_validator

from .jsonapi import Response
from .baseinfo import BaseInfoResource
from .entries import EntryInfoResource, EntryResource
from .index_metadb import IndexInfoResource
from .links import LinksResource
from .optimade_json import Success, ResponseMeta, OptimadeError
from .references import ReferenceResource
from .structures import StructureResource


__all__ = (
    "ErrorResponse",
    "EntryInfoResponse",
    "IndexInfoResponse",
    "InfoResponse",
    "LinksResponse",
    "EntryResponseOne",
    "EntryResponseMany",
    "StructureResponseOne",
    "StructureResponseMany",
    "ReferenceResponseOne",
    "ReferenceResponseMany",
)


class ErrorResponse(Response):
    """errors MUST be present and data MUST be skipped"""

    meta: Optional[ResponseMeta] = Field(
        None, description="A meta object containing non-standard information"
    )
    errors: List[OptimadeError] = Field(
        ...,
        description="A list of OPTIMADE-specific JSON API error objects, where the field detail MUST be present.",
        uniqueItems=True,
    )

    @root_validator(pre=True)
    def data_must_be_skipped(cls, values):
        if values.get("data", None) is not None:
            raise ValueError("data MUST be skipped for failures reporting errors")
        return values


class IndexInfoResponse(Success):
    data: IndexInfoResource = Field(..., description="Index meta-database /info data")


class EntryInfoResponse(Success):
    data: EntryInfoResource = Field(
        ..., description="OPTIMADE information for an entry endpoint"
    )


class InfoResponse(Success):
    data: BaseInfoResource = Field(..., description="The implementations /info data")


class EntryResponseOne(Success):
    data: Union[EntryResource, Dict[str, Any], None] = Field(...)
    included: Optional[Union[List[EntryResource], List[Dict[str, Any]]]] = Field(
        None, uniqueItems=True
    )


class EntryResponseMany(Success):
    data: Union[List[EntryResource], List[Dict[str, Any]]] = Field(
        ..., uniqueItems=True
    )
    included: Optional[Union[List[EntryResource], List[Dict[str, Any]]]] = Field(
        None, uniqueItems=True
    )


class LinksResponse(EntryResponseMany):
    data: Union[List[LinksResource], List[Dict[str, Any]]] = Field(
        ...,
        description="List of unique OPTIMADE links resource objects",
        uniqueItems=True,
    )


class StructureResponseOne(EntryResponseOne):
    data: Union[StructureResource, Dict[str, Any], None] = Field(
        ..., description="A single structures entry resource"
    )


class StructureResponseMany(EntryResponseMany):
    data: Union[List[StructureResource], List[Dict[str, Any]]] = Field(
        ...,
        description="List of unique OPTIMADE structures entry resource objects",
        uniqueItems=True,
    )


class ReferenceResponseOne(EntryResponseOne):
    data: Union[ReferenceResource, Dict[str, Any], None] = Field(
        ..., description="A single references entry resource"
    )


class ReferenceResponseMany(EntryResponseMany):
    data: Union[List[ReferenceResource], List[Dict[str, Any]]] = Field(
        ...,
        description="List of unique OPTIMADE references entry resource objects",
        uniqueItems=True,
    )



import operator
from optimade.filterparser import LarkParser
from lark import Tree
from lark.lexer import Token
from django.db.models import Q


class DjangoQueryError(Exception):
    pass


django_db_keys = {
    "natoms": "entry__natoms",
    "chemical_formula": "composition__formula__in",
    "stability": "stability",
    "band_gap": "calculation__band_gap",
    "element": "composition__element_list__contains",
    "nelements": "entry__composition__ntypes",
}


class Lark2Django:
    def __init__(self):
        self.opers = {
            "=": self.eq,
            ">": self.gt,
            ">=": self.ge,
            "<": self.lt,
            "<=": self.le,
            "!=": self.ne,
            "OR": self.or_,
            "AND": self.and_,
            "NOT": self.not_,
        }
        self.parser = LarkParser(version=(0, 9, 7))

    def parse_raw_q(self, raw_query):
        return self.parser.parse(raw_query)

    def eq(self, a, b):
        return Q(**{a: b})

    def gt(self, a, b):
        return Q(**{a + "__gt": b})

    def ge(self, a, b):
        return Q(**{a + "__gte": b})

    def lt(self, a, b):
        return Q(**{a + "__lt": b})

    def le(self, a, b):
        return Q(**{a + "__lte": b})

    def ne(self, a, b):
        return ~Q(**{a: b})

    def not_(self, a):
        return ~a

    def and_(self, a, b):
        return operator.and_(a, b)

    def or_(self, a, b):
        return operator.or_(a, b)

    def evaluate(self, parse_Tree):
        if isinstance(parse_Tree, Tree):
            children = parse_Tree.children
            if len(children) == 1:
                return self.evaluate(children[0])
            elif len(children) == 2:
                op_fn = self.evaluate(children[0])
                return op_fn(self.evaluate(children[1]))
            elif len(children) == 3:
                if parse_Tree.data == "comparison":
                    db_prop = self.evaluate(children[0])
                    op_fn = self.evaluate(children[1])

                    if db_prop in django_db_keys.keys():
                        return op_fn(
                            django_db_keys[db_prop], self.evaluate(children[2])
                        )
                    else:
                        raise DjangoQueryError(
                            "Unknown property is queried : " + (db_prop)
                        )

                else:
                    op_fn = self.evaluate(children[1])
                    return op_fn(self.evaluate(children[0]), self.evaluate(children[2]))
            else:
                raise DjangoQueryError("Not compatible format. Tree has >3 children")

        elif isinstance(parse_Tree, Token):
            if parse_Tree.type == "VALUE":
                return parse_Tree.value
            elif parse_Tree.type in ["NOT", "CONJUNCTION", "OPERATOR"]:
                return self.opers[parse_Tree.value]
        else:
            raise DjangoQueryError("Not a Lark Tree or Token")



# Copyright 2016-2018 Fawzi Mohamed, Lauri Himanen, Danio Brambila, Ankit Kariryaa, Henning Glawe
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import unittest
import logging

from vaspparser import VASPRunParserInterface


def get_result(folder, metaname=None):
    """Get the results from the calculation in the given folder. By default goes through different

    Args:
        folder: The folder relative to the directory of this script where the
            parsed calculation resides.
        metaname(str): Optional quantity to return. If not specified, returns
            the full dictionary of results.
    """
    dirname = os.path.dirname(__file__)
    filename = os.path.join(dirname, folder, "vasprun.xml.static")
    parser = VASPRunParserInterface(None, debug=True, log_level=logging.CRITICAL)
    results = parser.parse(filename)

    if metaname is None:
        return results
    else:
        return results[metaname]


class TestDOS(unittest.TestCase):
    """Tests that the parser can handle single point calculations.
    """
    @classmethod
    def setUpClass(cls):
        cls.results = get_result(os.path.join("examples", "dos"))

    def test_program_name(self):
        result = self.results["program_name"]
        self.assertEqual(result, "VASP")

    def test_program_basis_set_type(self):
        result = self.results["program_basis_set_type"]
        self.assertEqual(result, "plane waves")

    def test_electronic_structure_method(self):
        result = self.results["electronic_structure_method"]
        self.assertEqual(result, "DFT")

    def test_dos_energies_and_values(self):
        """Test that all the DOS related values are found and that the
        dimensions match.
        """
        dos_values = self.results["dos_values"]
        dos_energies = self.results["dos_energies"]
        dos_energies_normalized = self.results["dos_energies_normalized"]
        dos_integrated_values = self.results["dos_integrated_values"]

        self.assertEqual(len(dos_energies.shape), 1)
        self.assertEqual(len(dos_values.shape), 2)
        self.assertEqual(len(dos_integrated_values.shape), 2)
        self.assertEqual(dos_energies.shape[0], dos_values.shape[1])
        self.assertEqual(dos_energies_normalized.shape[0], dos_values.shape[1])


if __name__ == '__main__':
    suites = []
    suites.append(unittest.TestLoader().loadTestsFromTestCase(TestDOS))

    alltests = unittest.TestSuite(suites)
    unittest.TextTestRunner(verbosity=0).run(alltests)



# pylint: disable=relative-beyond-top-level,import-outside-toplevel
import os
import unittest
from traceback import print_exc

from optimade.validator import ImplementationValidator

from .utils import SetClient


class ServerTestWithValidator(SetClient, unittest.TestCase):

    server = "regular"

    def test_with_validator(self):
        validator = ImplementationValidator(client=self.client)
        try:
            validator.main()
        except Exception:
            print_exc()
        self.assertTrue(validator.valid)


class IndexServerTestWithValidator(SetClient, unittest.TestCase):

    server = "index"

    def test_with_validator(self):
        validator = ImplementationValidator(client=self.client, index=True)
        try:
            validator.main()
        except Exception:
            print_exc()
        self.assertTrue(validator.valid)


def test_mongo_backend_package_used():
    import pymongo
    import mongomock
    from optimade.server.entry_collections import client

    force_mongo_env_var = os.environ.get("OPTIMADE_CI_FORCE_MONGO", None)
    if force_mongo_env_var is None:
        return

    if int(force_mongo_env_var) == 1:
        assert issubclass(client.__class__, pymongo.MongoClient)
    elif int(force_mongo_env_var) == 0:
        assert issubclass(client.__class__, mongomock.MongoClient)
    else:
        raise Exception(
            "The environment variable OPTIMADE_CI_FORCE_MONGO cannot be parsed as an int."
        )


class AsTypeTestsWithValidator(SetClient, unittest.TestCase):

    server = "regular"

    def test_as_type_with_validator(self):

        test_urls = {
            "http://example.org/v0/structures": "structures",
            "http://example.org/v0/structures/mpf_1": "structure",
            "http://example.org/v0/references": "references",
            "http://example.org/v0/references/dijkstra1968": "reference",
            "http://example.org/v0/info": "info",
            "http://example.org/v0/links": "links",
        }
        with unittest.mock.patch(
            "requests.get", unittest.mock.Mock(side_effect=self.client.get)
        ):
            for url, as_type in test_urls.items():
                validator = ImplementationValidator(
                    base_url=url, as_type=as_type, verbosity=5
                )
                try:
                    validator.main()
                except Exception:
                    print_exc()
                self.assertTrue(validator.valid)



import os
import sys

try:
    from optimade import __version__
except ImportError:
    raise ImportError(
        "optimade needs to be installed prior to running 'assert_version.py'"
    )

package_version = f"v{__version__}"

tag_version = os.getenv("TAG_VERSION")
tag_version = tag_version[len("refs/tags/") :]

if tag_version == package_version:
    print(f"The versions match: tag:'{tag_version}' == package:'{package_version}'")
    sys.exit(0)

print(
    f"""The current package version '{package_version}' does not equal the tag version '{tag_version}'.
Update package version by \"invoke setver --new-ver='{tag_version[1:]}'\" and re-commit.
Please remove the tag from both GitHub and your local repository and try again!"""
)
sys.exit(1)



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
DFT specific metadata
'''

import re

from nomad import config, utils
from nomad.metainfo import MSection, Section, Quantity, MEnum, SubSection
from nomad.metainfo.search_extension import Search

from .optimade import OptimadeEntry
from .metainfo.public import Workflow, fast_access
from .metainfo.public import section_XC_functionals


xc_treatments = {
    'gga': 'GGA',
    'hf_': 'HF',
    'oep': 'OEP',
    'hyb': 'hybrid',
    'mgg': 'meta-GGA',
    'vdw': 'vdW',
    'lda': 'LDA',
}
''' https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-info/wikis/metainfo/XC-functional '''

basis_sets = {
    'gaussians': 'gaussians',
    'realspacegrid': 'real-space grid',
    'planewaves': 'plane waves'
}

compound_types = [
    'unary',
    'binary',
    'ternary',
    'quaternary',
    'quinary',
    'sexinary',
    'septenary',
    'octanary',
    'nonary',
    'decinary'
]

_electronic_quantities = [
    'electronic_band_structure',
    'electronic_dos',
    'eigenvalues_values',
]

_mechanical_quantities = [
    'stress_tensor'
]

_thermal_quantities = [
    'thermodynamical_property_heat_capacity_C_v',
    'vibrational_free_energy_at_constant_volume',
    'phonon_band_structure',
    'phonon_dos',
]

_magnetic_quantities = [
    'spin_S2'
]

_optical_quantities = [
    'oscillator_strengths',
    'transition_dipole_moments'
]

_searchable_quantities = set(_electronic_quantities + _mechanical_quantities + _thermal_quantities + _magnetic_quantities + _optical_quantities)

version_re = re.compile(r'(\d+(\.\d+(\.\d+)?)?)')


def map_functional_name_to_xc_treatment(name):
    if name == config.services.unavailable_value:
        return name

    return xc_treatments.get(name[:3].lower(), config.services.unavailable_value)


def map_basis_set_to_basis_set_label(name):
    key = name.replace('_', '').replace('-', '').replace(' ', '').lower()
    return basis_sets.get(key, name)


def simplify_version(version):
    match = version_re.search(version)
    if match is None:
        return version
    else:
        return match.group(0)


class Label(MSection):
    '''
    Label that further classify a structure.

    Attributes:
        label: The label as a string
        type: The type of the label
        source: The source that this label was taken from.

    '''
    label = Quantity(type=str, a_search=Search())

    type = Quantity(type=MEnum(
        'compound_class', 'classification', 'prototype', 'prototype_id'),
        a_search=Search())

    source = Quantity(
        type=MEnum('springer', 'aflow_prototype_library'),
        a_search=Search())


class DFTMetadata(MSection):
    m_def = Section(a_domain='dft')

    basis_set = Quantity(
        type=str, default='not processed',
        description='The used basis set functions.',
        a_search=Search(statistic_values=[
            '(L)APW+lo', 'gaussians', 'numeric AOs', 'plane waves', 'psinc functions',
            'real-space grid', 'unavailable', 'not processed'
        ]))

    xc_functional = Quantity(
        type=str, default='not processed',
        description='The libXC based xc functional classification used in the simulation.',
        a_search=Search(
            statistic_values=list(xc_treatments.values()) + ['unavailable', 'not processed'],
            statistic_size=100))

    xc_functional_names = Quantity(
        type=str, default=[], shape=['*'],
        description='The list of libXC functional names that where used in this entry.',
        a_search=Search(many_and='append'))

    system = Quantity(
        type=str, default='not processed',
        description='The system type of the simulated system.',
        a_search=Search(statistic_values=[
            '1D', '2D', 'atom', 'bulk', 'molecule / cluster', 'surface',
            'unavailable', 'not processed'
        ]))

    compound_type = Quantity(
        type=str, default='not processed',
        description='The compound type of the simulated system.',
        a_search=Search(statistic_values=compound_types + ['not processed'])
    )

    crystal_system = Quantity(
        type=str, default='not processed',
        description='The crystal system type of the simulated system.',
        a_search=Search(
            statistic_values=[
                'cubic', 'hexagonal', 'monoclinic', 'orthorhombic', 'tetragonal',
                'triclinic', 'trigonal', 'unavailable', 'not processed']
        ))

    spacegroup = Quantity(
        type=int, default=-1,
        description='The spacegroup of the simulated system as number.',
        a_search=Search())

    spacegroup_symbol = Quantity(
        type=str, default='not processed',
        description='The spacegroup as international short symbol.',
        a_search=Search())

    code_name = Quantity(
        type=str, default='not processed',
        description='The name of the used code.',
        a_search=Search())  # in import the parser module is added codes here as statistic_values

    code_version = Quantity(
        type=str, default='not processed',
        description='The version of the used code.',
        a_search=Search())

    n_geometries = Quantity(
        type=int, default=0, description='Number of unique geometries.',
        a_search=Search(metric_name='geometries', metric='sum'))

    n_calculations = Quantity(
        type=int, default=0,
        description='Number of single configuration calculation sections',
        a_search=Search(metric_name='calculations', metric='sum'))

    n_total_energies = Quantity(
        type=int, default=0, description='Number of total energy calculations',
        a_search=Search(metric_name='total_energies', metric='sum'))

    n_quantities = Quantity(
        type=int, default=0, description='Number of metainfo quantities parsed from the entry.',
        a_search=Search(metric='sum', metric_name='quantities'))

    quantities = Quantity(
        type=str, shape=['0..*'],
        description='All quantities that are used by this entry.',
        a_search=Search(
            metric_name='distinct_quantities', metric='cardinality', many_and='append'))

    searchable_quantities = Quantity(
        type=str, shape=['0..*'],
        description='All quantities with existence filters in the search GUI.',
        a_search=Search(many_and='append', statistic_size=len(_searchable_quantities) + 15))  # Temporarily increased the statistics size while migrating from old set to new one.

    geometries = Quantity(
        type=str, shape=['0..*'],
        description='Hashes for each simulated geometry',
        a_search=Search(metric_name='unique_geometries', metric='cardinality'))

    group_hash = Quantity(
        type=str,
        description='Hashes that describe unique geometries simulated by this code run.',
        a_search=Search(many_or='append', group='groups_grouped', metric_name='groups', metric='cardinality'))

    labels = SubSection(
        sub_section=Label, repeats=True, categories=[fast_access],
        description='The labels taken from AFLOW prototypes and springer.',
        a_search='labels')

    labels_springer_compound_class = Quantity(
        type=str, shape=['0..*'],
        description='Springer compund classification.',
        a_search=Search(
            many_and='append', statistic_size=20,
            statistic_order='_count'))

    labels_springer_classification = Quantity(
        type=str, shape=['0..*'],
        description='Springer classification by property.',
        a_search=Search(
            many_and='append', statistic_size=20,
            statistic_order='_count'))

    optimade = SubSection(
        sub_section=OptimadeEntry,
        description='Metadata used for the optimade API.',
        a_search='optimade')

    workflow = Quantity(type=Workflow, a_search=Search())

    def code_name_from_parser(self):
        entry = self.m_parent
        if entry.parser_name is not None:
            from nomad.parsing.parsers import parser_dict
            parser = parser_dict.get(entry.parser_name)
            if hasattr(parser, 'code_name'):
                return parser.code_name
        return config.services.unavailable_value

    def update_group_hash(self):
        user_id = None
        uploader = self.m_parent.uploader
        if uploader is not None:
            user_id = uploader.user_id
        self.group_hash = utils.hash(
            self.m_parent.formula,
            self.spacegroup,
            self.basis_set,
            self.xc_functional,
            self.code_name,
            self.code_version,
            self.m_parent.with_embargo,
            user_id)

    def apply_domain_metadata(self, entry_archive):
        from nomad.normalizing.system import normalized_atom_labels
        entry = self.m_parent

        logger = utils.get_logger(__name__).bind(
            upload_id=entry.upload_id, calc_id=entry.calc_id, mainfile=entry.mainfile)

        self.code_name = self.code_name_from_parser()

        if entry_archive is None:
            return

        section_run = entry_archive.section_run
        if not section_run:
            logger.warn('no section_run found')
            return
        section_run = section_run[0]

        # default values
        self.system = config.services.unavailable_value
        self.crystal_system = config.services.unavailable_value
        self.spacegroup_symbol = config.services.unavailable_value
        self.basis_set = config.services.unavailable_value
        self.xc_functional = config.services.unavailable_value

        section_system = None
        for section in section_run.section_system:
            if section.is_representative:
                section_system = section
                break

        # code and code specific ids
        try:
            code_name = section_run.program_name
            if code_name:
                self.code_name = code_name
            else:
                raise KeyError
        except KeyError as e:
            logger.warn('archive without program_name', exc_info=e)

        try:
            version = section_run.program_version
            if version:
                self.code_version = simplify_version(version)
            else:
                raise KeyError
        except KeyError:
            self.code_version = config.services.unavailable_value

        def get_value(value):
            return value if value else config.services.unavailable_value

        raw_id = section_run.raw_id
        if raw_id is not None:
            entry.raw_id = raw_id

        # metadata (system, method, chemistry)
        atom_labels = section_system.atom_labels if section_system else []
        atoms = atom_labels if atom_labels else []
        entry.n_atoms = len(atoms)
        atoms = list(set(normalized_atom_labels(set(atoms))))
        atoms.sort()
        entry.atoms = atoms
        self.compound_type = compound_types[len(atoms) - 1] if len(atoms) <= 10 else '>decinary'

        self.system = config.services.unavailable_value
        self.crystal_system = config.services.unavailable_value
        self.spacegroup_symbol = config.services.unavailable_value

        section_symmetry = None
        if section_system and len(section_system.section_symmetry) > 0:
            section_symmetry = section_system.section_symmetry[0]
            self.crystal_system = get_value(section_symmetry.crystal_system)
            spacegroup = section_symmetry.space_group_number
            self.spacegroup = 0 if not spacegroup else int(spacegroup)
            self.spacegroup_symbol = get_value(section_symmetry.international_short_symbol)

        program_basis_set_type = section_run.program_basis_set_type
        if program_basis_set_type:
            self.basis_set = map_basis_set_to_basis_set_label(program_basis_set_type)

        if section_system:
            self.system = get_value(section_system.system_type)
            entry.formula = get_value(section_system.chemical_composition_bulk_reduced)

        # metrics and quantities
        quantities = set()
        searchable_quantities = set()
        geometries = set()
        xc_functionals = set()
        xc_functional = None

        n_quantities = 0
        n_calculations = 0
        n_total_energies = 0
        n_geometries = 0

        for section, property_def, _ in entry_archive.m_traverse():
            property_name = property_def.name
            quantities.add(property_name)
            n_quantities += 1

            if property_name in _searchable_quantities:
                searchable_quantities.add(property_name)

            if property_def == section_XC_functionals.XC_functional_name:
                xc_functional = getattr(section, property_name)
                if xc_functional:
                    xc_functionals.add(xc_functional)

            if property_name == 'energy_total':
                n_total_energies += 1

            if property_name == 'configuration_raw_gid':
                geometries.add(section.m_get(property_def))

            if property_name == 'section_single_configuration_calculation':
                n_calculations += 1

            if property_name == 'section_system':
                n_geometries += 1

        self.xc_functional_names = sorted(xc_functionals)
        if len(self.xc_functional_names) > 0:
            self.xc_functional = map_functional_name_to_xc_treatment(
                get_value(self.xc_functional_names[0]))
        else:
            self.xc_functional = config.services.unavailable_value

        self.quantities = list(quantities)
        self.geometries = list(geometries)
        self.searchable_quantities = list(searchable_quantities)
        self.n_quantities = n_quantities
        self.n_calculations = n_calculations
        self.n_total_energies = n_total_energies
        self.n_geometries = n_geometries

        # grouping
        self.update_group_hash()

        # labels
        compounds = set()
        classifications = set()
        if section_system:
            for section in section_system.section_springer_material:
                compounds.update(section.springer_compound_class)
                classifications.update(section.springer_classification)

        for compound in compounds:
            self.labels.append(Label(label=compound, type='compound_class', source='springer'))
        for classification in classifications:
            self.labels.append(Label(label=classification, type='classification', source='springer'))
        self.labels_springer_compound_class = list(compounds)
        self.labels_springer_classification = list(classifications)

        aflow_id, aflow_label = None, None
        section_prototype = section_system.section_prototype if section_system else []
        if section_prototype:
            aflow_id = get_value(section_prototype[0].prototype_aflow_id)
            aflow_label = get_value(section_prototype[0].prototype_label)

        if aflow_id is not None and aflow_label is not None:
            self.labels.append(Label(label=aflow_label, type='prototype', source='aflow_prototype_library'))
            self.labels.append(Label(label=aflow_id, type='prototype_id', source='aflow_prototype_library'))

        if entry_archive.section_workflow:
            self.workflow = entry_archive.section_workflow



import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from vaspparser.metainfo import vasp_incars
from nomad.datamodel.metainfo import public
from vaspparser.metainfo import vasp_incarsOut

m_package = Package(
    name='vasp_incarsUnknown_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='vasp.incarsUnknown.nomadmetainfo.json'))


class x_vasp_incarUnknown_param(MCategory):
    '''
    Unknown incar parameters.  By 'unknown' we refer to incar parameters in the OLD
    (predated Sept.2019) 'vasp.nomadmetainfo.json' that are **not** listed in the current
    (Oct.2019) VaspWiki
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_vasp_incarUnknown_param'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_vasp_incarOut_ENMAX = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Maximum cutoff (normally specified only in POTCAR). Value prinded out after
        evaluating the input.
        ''',
        categories=[vasp_incarsOut.x_vasp_incarOut_param],
        a_legacy=LegacyDefinition(name='x_vasp_incarOut_ENMAX'))

    x_vasp_incar_ENMAX = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Maximum cutoff (normally specified only in POTCAR). Value prinded out after
        evaluating the input.
        ''',
        categories=[vasp_incars.x_vasp_incar_param],
        a_legacy=LegacyDefinition(name='x_vasp_incar_ENMAX'))

    x_vasp_incarOut_LCOMPAT = Quantity(
        type=bool,
        shape=[],
        description='''
        In vasp.4.2 the augmentation charges are forced to be zero at the boundary of the
        augmentation sphere, therefore results are slightly different from vasp.3.2
        (usually differences are smaller than 0.01 meV). The old behavior can be restored
        by setting LCOMPAT = .TRUE. in the INCAR file.
        ''',
        categories=[vasp_incarsOut.x_vasp_incarOut_param],
        a_legacy=LegacyDefinition(name='x_vasp_incarOut_LCOMPAT'))


m_package.__init_metainfo__()



import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='openkim_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='openkim.nomadmetainfo.json'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    openkim_build_date = Quantity(
        type=str,
        shape=[],
        description='''
        build date as string
        ''',
        categories=[public.accessory_info, public.program_info],
        a_legacy=LegacyDefinition(name='openkim_build_date'))

    openkim_src_date = Quantity(
        type=str,
        shape=[],
        description='''
        date of last modification of the source as string
        ''',
        categories=[public.accessory_info, public.program_info],
        a_legacy=LegacyDefinition(name='openkim_src_date'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_openkim_atom_kind_refs = Quantity(
        type=public.section_method_atom_kind,
        shape=['number_of_atoms'],
        description='''
        reference to the atom kinds of each atom
        ''',
        a_legacy=LegacyDefinition(name='x_openkim_atom_kind_refs'))


m_package.__init_metainfo__()
